import base64
import dataclasses
import json
import typing
from typing import Generic, TypeVar, Dict

from super_cereal.cerealizer import Cerealizer, T

E = TypeVar('E')


@dataclasses.dataclass
class Encrypted(Generic[E]):
    """
    Container class for automatically encrypting during serialization and decrypting during deserialization.

    `value` will be None when initially set as None, or when the encryption key for `key_id` is not present.
    """
    key_id: str
    value: typing.Optional[E]


class EncryptedCerealizer(Cerealizer[Encrypted[E], Dict[str, E]]):

    def __init__(self, keys: Dict[str, bytes], value_cerealizer: Cerealizer) -> None:
        super().__init__(keys)
        self.value_cerealizer = value_cerealizer

    @staticmethod
    def enabled() -> bool:
        try:
            from Crypto.Cipher import AES
            return True
        except ImportError:
            return False

    def serialize(self, obj: Encrypted[E], t: T = None) -> Dict[str, E]:
        """
        https://pycryptodome.readthedocs.io/en/latest/src/cipher/modern.html?highlight=gcm#gcm-mode
        """
        from Crypto.Cipher import AES

        if t:
            t = typing.get_args(obj)
        if not t:
            t = type(obj.value)

        cipher = AES.new(self.encryption_keys[obj.key_id], AES.MODE_GCM)
        ciphertext, tag = cipher.encrypt_and_digest(json.dumps(self.value_cerealizer.serialize(obj.value, t)).encode())

        return {
            'key_id': obj.key_id,
            'value': base64.b64encode(ciphertext).decode(),
            'tag': base64.b64encode(tag).decode(),
            'nonce': base64.b64encode(cipher.nonce).decode()
        }

    def deserialize(self, obj: Dict[str, E], t: Encrypted[E]) -> Encrypted[E]:
        from Crypto.Cipher import AES

        t = typing.get_args(t)[0]

        try:
            key_id = self.encryption_keys[obj['key_id']]
        except KeyError:
            return Encrypted(obj['key_id'], None)

        cipher = AES.new(key_id, AES.MODE_GCM, nonce=base64.b64decode(obj['nonce']))
        plaintext = cipher.decrypt_and_verify(base64.b64decode(obj['value']), base64.b64decode(obj['tag']))

        payload = json.loads(plaintext)
        return Encrypted(key_id=obj['key_id'], value=self.value_cerealizer.deserialize(payload, t))



import abc
import typing
from abc import ABC
from typing import TypeVar, Generic, Dict, Optional

T = TypeVar('T')
V = TypeVar('V')


class SerializationException(Exception):
    pass


class DeserializationException(Exception):
    pass


class ICerealizer(Generic[T, V], abc.ABC):  # pragma: no cover

    @abc.abstractmethod
    def serialize(self, obj: any, t: T = None) -> V:
        pass

    @abc.abstractmethod
    def deserialize(self, obj: V, t: T) -> T:
        pass


class ITypeRegistry(Dict[type, ICerealizer], abc.ABC):
    @property
    @abc.abstractmethod
    def default(self):
        return self.default


class Cerealizer(ICerealizer[T, V], ABC):
    def __init__(self, encryption_keys: Dict[str, bytes] = None):
        if encryption_keys is None:
            encryption_keys = {}
        self.encryption_keys = encryption_keys

    def add_registry(self, registry: Optional[ITypeRegistry]):
        self.registry = registry


class TheTypeRegistry(ITypeRegistry):
    def __init__(self) -> None:
        super().__init__()
        self.d: Dict[type, Cerealizer] = {}
        self._default = None

    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, value):
        if self._default is not None:
            self._default.add_registry(None)

        self._default = value

        value.add_registry(self)

    def __contains__(self, item):
        return item in self.d

    def __getitem__(self, item):
        if item in self.d:
            return self.d[item]
        if typing.get_origin(item) in self.d:
            return self.d[typing.get_origin(item)]

        return self.default

    def __setitem__(self, key: type, value: Cerealizer):
        self.d[key] = value
        value.add_registry(self)

    def __delitem__(self, key):
        popped = self.d.pop(key)
        popped.add_registry(None)

    def __len__(self):
        return len(self.d) + 1



import json
import typing

from super_cereal.cerealizer import Cerealizer, T, TheTypeRegistry
from super_cereal.cerealizer.builtins import PassthruCerealizer, ListCerealizer, UnionCerealizer, DictCerealizer
from super_cereal.cerealizer.encryption import EncryptedCerealizer, Encrypted

JsonTypes = typing.Union[str, float, int, bool, type(None), list, dict]


class JsonCerealizer(Cerealizer[T, JsonTypes]):
    def __init__(self, encryption_keys: typing.Dict[str, bytes] = None):
        registry = TheTypeRegistry()
        registry[type(None)] = PassthruCerealizer()
        registry[str] = PassthruCerealizer()
        registry[float] = PassthruCerealizer()
        registry[int] = PassthruCerealizer()
        registry[bool] = PassthruCerealizer()
        registry[list] = ListCerealizer()
        registry[typing.List] = ListCerealizer()
        registry[typing.Union] = UnionCerealizer()

        if EncryptedCerealizer.enabled():
            registry[Encrypted] = EncryptedCerealizer(encryption_keys, self)

        registry.default = DictCerealizer()

        self.registry = registry

    def serialize(self, obj: any, expected_type: T = None) -> JsonTypes:
        if expected_type is None:
            expected_type = type(obj)

        if expected_type in self.registry:
            return self.registry[expected_type].serialize(obj, expected_type)
        if typing.get_origin(expected_type) in self.registry:
            return self.registry[typing.get_origin(expected_type)].serialize(obj, expected_type)

        return self.registry.default.serialize(obj, expected_type)

    def deserialize(self, obj: JsonTypes, t: T) -> T:
        if t in self.registry:
            return self.registry[t].deserialize(obj, t)
        if typing.get_origin(t) in self.registry:
            return self.registry[typing.get_origin(t)].deserialize(obj, t)

        return self.registry.default.deserialize(obj, t)


class JsonByteCerealizer(JsonCerealizer):

    def __init__(self, encryption_keys: typing.Dict[str, bytes] = None):
        super().__init__(encryption_keys)

    def serialize(self, obj: any, expected_type: T = None) -> bytes:
        return json.dumps(super().serialize(obj, expected_type)).encode()

    def deserialize(self, obj: bytes, t: T) -> T:
        return super().deserialize(json.loads(obj.decode()), t)



import inspect
import typing

from super_cereal.cerealizer import Cerealizer, T, V, SerializationException, DeserializationException


class PassthruCerealizer(Cerealizer[any, any]):
    def serialize(self, obj: any, t: any = None) -> any:
        return None if obj is None else t(obj)

    def deserialize(self, obj: any, t: any) -> any:
        return None if obj is None else t(obj)


class ListCerealizer(Cerealizer):
    def serialize(self, obj: any, t: T = None) -> V:
        t = typing.get_args(t)[0]
        return [self.registry[t].serialize(v, t) for v in obj]

    def deserialize(self, obj: V, t: T) -> T:
        t = typing.get_args(t)[0]
        return [self.registry[t].deserialize(v, t) for v in obj]


class UnionCerealizer(Cerealizer):
    def serialize(self, obj: any, t: T = None) -> V:
        for t1 in typing.get_args(t):
            if type(obj) == t1:
                return self.registry[t1].serialize(obj, t1)
            if type(obj) == typing.get_origin(t1):
                return self.registry[t1].serialize(obj, t1)

        raise NotImplementedError()

    def deserialize(self, obj: V, t: T) -> T:
        args = typing.get_args(t)
        for arg in args:
            if arg == type(obj):
                return self.registry[arg].deserialize(obj, arg)
            if typing.get_origin(arg) == type(obj):
                return self.registry[arg].deserialize(obj, arg)

        raise NotImplementedError()


class DictCerealizer(Cerealizer):
    def serialize(self, obj: any, t: T = None) -> V:
        if t == dict or typing.get_origin(t) == dict:
            return obj

        # noinspection PyTypeChecker
        fields: typing.List[typing.Tuple[str, inspect.Parameter]] = list(
            inspect.signature(t.__init__).parameters.items())[1:]

        for field, param in fields:
            # noinspection PyUnresolvedReferences,PyProtectedMember
            if param.annotation == inspect._empty:
                class_name = f'"{t.__module__}.{t.__name__}"'
                raise SerializationException(f'{class_name}: "{field}" has no annotation.')

        fixed_fields = {}

        for field, param in fields:
            cerealizer = self.registry[param.annotation]
            fixed_fields[field] = cerealizer.serialize(getattr(obj, field), param.annotation)

        return fixed_fields

    def deserialize(self, obj: V, t: T) -> T:
        if t == dict or typing.get_origin(t) == dict:
            return obj

        # noinspection PyTypeChecker
        fields: typing.List[
            typing.Tuple[str, inspect.Parameter]] = list(inspect.signature(t.__init__).parameters.items())[1:]

        for field, param in fields:
            # noinspection PyUnresolvedReferences,PyProtectedMember
            if param.annotation == inspect._empty:
                class_name = f'"{t.__module__}.{t.__name__}"'
                raise DeserializationException(f'{class_name}: "{field}" has no annotation.')

        fixed = {
            field: self.registry[param.annotation].deserialize(obj[field], param.annotation)
            for field, param in fields
        }

        return t(**fixed)



import inspect
import io
import json
from typing import Union, get_origin, get_args, Dict, List, Tuple

from avro.datafile import DataFileReader
from avro.io import DatumReader

from super_cereal.cerealizer import Cerealizer, V, T
from super_cereal.cerealizer.encryption import Encrypted
from super_cereal.cerealizer.json import JsonCerealizer

BUILTIN_ALIASES = {
    str: 'string',
    int: 'int',
    bytes: 'bytes',
    float: 'double',
    bool: 'boolean',
    type(None): 'null',
}


class AvroCerealizer(Cerealizer):
    def __init__(self, encryption_keys: Dict[str, bytes] = None) -> None:
        super().__init__(encryption_keys)
        self.json_serializer = JsonCerealizer(encryption_keys)

    @staticmethod
    def get_schema(record: type) -> Dict[str, any]:
        def get_schema(t: type, namespace: str) -> Dict[str, any]:
            if t in BUILTIN_ALIASES:
                return {'type': BUILTIN_ALIASES[t]}
            if t == list or get_origin(t) == list:
                arg = get_args(t)[0]
                if arg in BUILTIN_ALIASES:
                    return {'type': 'array', 'items': BUILTIN_ALIASES[arg]}
                return {'type': 'array', 'items': get_schema(arg, namespace)}
            if Union == get_origin(t):
                args = get_args(t)

                def type_for_arg(a):
                    return BUILTIN_ALIASES[a] if a in BUILTIN_ALIASES else get_schema(a, namespace)

                return {'type': [type_for_arg(z) for z in args]}
            if Encrypted == get_origin(t):
                print('stuff')
                return {
                    'namespace': namespace,
                    'type': 'record',
                    'name': get_origin(t).__name__,
                    'fields': [
                        {'name': 'key_id', 'type': 'string'},
                        {'name': 'value',
                         'type': [get_schema(get_args(t)[0], f'{namespace}.{get_origin(t).__name__}.value'), 'string']},
                        {'name': 'tag', 'type': 'string'},
                        {'name': 'nonce', 'type': 'string'},
                    ]
                }

            # noinspection PyTypeChecker
            fields: List[Tuple[str, inspect.Parameter]] = list(inspect.signature(t.__init__).parameters.items())[1:]

            f = []

            for field in fields:
                def should_simplify(tp):
                    return tp in BUILTIN_ALIASES or Union == get_origin(tp)

                schema = get_schema(field[1].annotation, namespace=f'{namespace}.{t.__name__}.{field[0]}')
                f.append({
                    'name': field[0],
                    'type': schema['type'] if should_simplify(field[1].annotation) else schema
                })

            return {
                'namespace': f'{inspect.getmodule(t).__name__}{namespace}',
                'type': 'record',
                'name': t.__name__,
                'fields': f
            }

        return get_schema(record, inspect.getmodule(record).__name__)

    def serialize(self, obj: any, t: T = None) -> V:
        import avro.schema
        from avro.datafile import DataFileWriter
        from avro.io import DatumWriter

        the_json = self.get_schema(type(obj))
        schema_json = json.dumps(the_json)
        schema = avro.schema.parse(schema_json)

        with io.BytesIO() as bytes_io:
            with DataFileWriter(bytes_io, DatumWriter(), schema) as writer:
                writer.append(self.json_serializer.serialize(obj))
                writer.flush()
                return bytes_io.getvalue()

    def deserialize(self, obj: bytes, t: T) -> T:
        with io.BytesIO(obj) as bytes_io:
            with DataFileReader(bytes_io, DatumReader()) as reader:
                for msg in reader:
                    return self.json_serializer.deserialize(msg, t)

        return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['super_cereal', 'super_cereal.cerealizer']

package_data = \
{'': ['*']}

extras_require = \
{'avro': ['avro>=1.11.1,<2.0.0'], 'encryption': ['pycryptodome>=3.16.0,<4.0.0']}

setup_kwargs = {
    'name': 'super-cereal',
    'version': '0.1.3',
    'description': 'Serialize Python objects',
    'long_description': '',
    'author': 'Sean Freitag',
    'author_email': 'sean.freitag@avant.com',
    'maintainer': 'Sean Freitag',
    'maintainer_email': 'sean.freitag@avant.com',
    'url': 'https://github.com/cowboygneox/super-cereal',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
